@@ -1,5 +1,6 @@
 use crate::{PreKeyBundle, X3DHError, X3dhVault, CSUITE};
 use ockam_core::lib::convert::TryFrom;
+use ockam_core::Result;
 use ockam_key_exchange_core::{CompletedKeyExchange, KeyExchanger};
 use ockam_vault_core::{
     Secret, SecretAttributes, SecretPersistence, SecretType, AES256_SECRET_LENGTH,
@@ -16,33 +17,36 @@ enum InitiatorState {
 /// The responder of X3DH receives a prekey bundle and computes the shared secret
 /// to communicate the first message to the initiator
 pub struct Initiator<V: X3dhVault> {
+    identity_key: Option<Secret>,
     ephemeral_identity_key: Option<Secret>,
     prekey_bundle: Option<PreKeyBundle>,
     state: InitiatorState,
     vault: V,
     completed_key_exchange: Option<CompletedKeyExchange>,
-    identity_key: Option<Secret>,
 }
 
 impl<V: X3dhVault> Initiator<V> {
     pub(crate) fn new(vault: V, identity_key: Option<Secret>) -> Self {
         Self {
+            identity_key,
             ephemeral_identity_key: None,
             prekey_bundle: None,
             state: InitiatorState::GenerateEphemeralIdentityKey,
             vault,
             completed_key_exchange: None,
-            identity_key,
         }
     }
 
     fn prologue(&mut self) -> ockam_core::Result<()> {
-        let p_atts = SecretAttributes::new(
-            SecretType::Curve25519,
-            SecretPersistence::Persistent,
-            CURVE25519_SECRET_LENGTH,
-        );
-        self.ephemeral_identity_key = Some(self.vault.secret_generate(p_atts)?);
+        if self.identity_key.is_none() {
+            let p_atts = SecretAttributes::new(
+                SecretType::Curve25519,
+                SecretPersistence::Persistent,
+                CURVE25519_SECRET_LENGTH,
+            );
+
+            self.identity_key = Some(self.vault.secret_generate(p_atts)?);
+        }
         Ok(())
     }
 }
@@ -62,22 +66,39 @@ impl<V: X3dhVault> std::fmt::Debug for Initiator<V> {
 }
 
 impl<V: X3dhVault> KeyExchanger for Initiator<V> {
-    fn process(&mut self, data: &[u8]) -> ockam_core::Result<Vec<u8>> {
+    fn generate_request(&mut self, _payload: &[u8]) -> Result<Vec<u8>> {
         match self.state {
             InitiatorState::GenerateEphemeralIdentityKey => {
                 self.prologue()?;
+                let identity_key = self.identity_key.as_ref().ok_or(X3DHError::InvalidState)?;
+                let pubkey = self.vault.secret_public_key_get(identity_key)?;
+
                 let ephemeral_identity_key = self.vault.secret_generate(SecretAttributes::new(
                     SecretType::Curve25519,
-                    SecretPersistence::Persistent,
+                    SecretPersistence::Ephemeral,
                     CURVE25519_SECRET_LENGTH,
                 ))?;
-                let pubkey = self.vault.secret_public_key_get(&ephemeral_identity_key)?;
+                let ephemeral_pubkey = self.vault.secret_public_key_get(&ephemeral_identity_key)?;
                 self.ephemeral_identity_key = Some(ephemeral_identity_key);
                 self.state = InitiatorState::ProcessPreKeyBundle;
-                Ok(pubkey.as_ref().to_vec())
+
+                let mut response = Vec::new();
+                response.extend_from_slice(pubkey.as_ref());
+                response.extend_from_slice(ephemeral_pubkey.as_ref());
+                Ok(response)
             }
+            InitiatorState::ProcessPreKeyBundle | InitiatorState::Done => {
+                Err(X3DHError::InvalidState.into())
+            }
+        }
+    }
+
+    fn handle_response(&mut self, response: &[u8]) -> Result<Vec<u8>> {
+        match self.state {
             InitiatorState::ProcessPreKeyBundle => {
-                let prekey_bundle = PreKeyBundle::try_from(data)?;
+                let prekey_bundle = PreKeyBundle::try_from(response)?;
+
+                let identity_key = self.identity_key.as_ref().ok_or(X3DHError::InvalidState)?;
 
                 let ephemeral_identity_key = self
                     .ephemeral_identity_key
@@ -90,25 +111,20 @@ impl<V: X3dhVault> KeyExchanger for Initiator<V> {
                     &prekey_bundle.identity_key,
                     prekey_bundle.signed_prekey.as_ref(),
                 )?;
-                let atts = SecretAttributes::new(
-                    SecretType::Curve25519,
-                    SecretPersistence::Ephemeral,
-                    CURVE25519_SECRET_LENGTH,
-                );
-                let esk = self.vault.secret_generate(atts)?;
+
                 let dh1 = self
                     .vault
-                    .ec_diffie_hellman(ephemeral_identity_key, &prekey_bundle.signed_prekey)?;
+                    .ec_diffie_hellman(identity_key, &prekey_bundle.signed_prekey)?;
                 let dh2 = self
                     .vault
-                    .ec_diffie_hellman(&esk, &prekey_bundle.identity_key)?;
+                    .ec_diffie_hellman(ephemeral_identity_key, &prekey_bundle.identity_key)?;
                 let dh3 = self
                     .vault
-                    .ec_diffie_hellman(&esk, &prekey_bundle.signed_prekey)?;
+                    .ec_diffie_hellman(ephemeral_identity_key, &prekey_bundle.signed_prekey)?;
                 let dh4 = self
                     .vault
-                    .ec_diffie_hellman(&esk, &prekey_bundle.one_time_prekey)?;
-                let mut ikm_bytes = vec![0xFFu8; 32];
+                    .ec_diffie_hellman(ephemeral_identity_key, &prekey_bundle.one_time_prekey)?;
+                let mut ikm_bytes = vec![0xFFu8; 32]; // FIXME: Why is it here?
                 ikm_bytes.extend_from_slice(self.vault.secret_export(&dh1)?.as_ref());
                 ikm_bytes.extend_from_slice(self.vault.secret_export(&dh2)?.as_ref());
                 ikm_bytes.extend_from_slice(self.vault.secret_export(&dh3)?.as_ref());
@@ -138,54 +154,22 @@ impl<V: X3dhVault> KeyExchanger for Initiator<V> {
                         .hkdf_sha256(&salt, CSUITE, Some(&ikm), vec![atts, atts])?;
                 let encrypt_key = keyrefs.pop().ok_or(X3DHError::InvalidState)?;
                 let decrypt_key = keyrefs.pop().ok_or(X3DHError::InvalidState)?;
-                let ek = self.vault.secret_public_key_get(&esk)?;
-                let pubkey = self.vault.secret_public_key_get(ephemeral_identity_key)?;
 
                 let mut state_hash = self.vault.sha256(CSUITE)?.to_vec();
                 state_hash.append(&mut ikm_bytes);
                 let state_hash = self.vault.sha256(state_hash.as_slice())?;
 
-                let mut aad = ek.as_ref().to_vec();
-                aad.extend_from_slice(&self.vault.sha256(pubkey.as_ref())?);
-                aad.extend_from_slice(CSUITE);
-                aad.extend_from_slice(&state_hash);
-
-                let atts = SecretAttributes::new(
-                    SecretType::Curve25519,
-                    SecretPersistence::Persistent,
-                    CURVE25519_SECRET_LENGTH,
-                );
-
-                let skb = if let Some(ik) = &self.identity_key {
-                    ik.clone()
-                } else {
-                    self.vault.secret_generate(atts)?
-                };
-                let ikb = self.vault.secret_public_key_get(&skb)?;
-
-                let mut plaintext = ikb.as_ref().to_vec();
-                plaintext
-                    .extend_from_slice(&self.vault.sign(ephemeral_identity_key, ikb.as_ref())?);
-
-                let mut ciphertext_and_tag = self.vault.aead_aes_gcm_encrypt(
-                    &encrypt_key,
-                    plaintext.as_slice(),
-                    &ek.as_ref()[..12],
-                    aad.as_slice(),
-                )?;
-                let mut output = aad[..64].to_vec();
-                output.append(&mut ciphertext_and_tag);
                 self.completed_key_exchange = Some(CompletedKeyExchange::new(
                     state_hash,
                     encrypt_key,
                     decrypt_key,
-                    skb,
-                    prekey_bundle.identity_key,
                 ));
                 self.state = InitiatorState::Done;
-                Ok(output)
+                Ok(vec![])
+            }
+            InitiatorState::GenerateEphemeralIdentityKey | InitiatorState::Done => {
+                Err(X3DHError::InvalidState.into())
             }
-            InitiatorState::Done => Err(X3DHError::InvalidState.into()),
         }
     }
 
